# test Benford's Law compliance

asc <- function(x) { as.character(x) }

flist <- system("ls *.csv", intern=TRUE);
dlist <- list();
for (jf in flist) {
  tpipe <- pipe(paste("awk 'NR>2'",jf));
  dlist[[jf]] <- read.csv(tpipe);
  print(names(dlist[[jf]]));
}
nlines <- 0;
for (jf in flist) {
  nlines <- nlines + dim(dlist[[jf]])[1];
}
mat <- matrix(NA, nlines, dim(dlist[[jf]])[2]-1);
precinct <- vector();
ii <- 1;
for (jf in flist) {
  precinct <- c(precinct,asc(dlist[[jf]]$Precinct));
  jlines <- dim(dlist[[jf]])[1];
  mat[ii:(ii+jlines-1),] <- as.matrix(dlist[[jf]][,-1]);
  ii <- ii + jlines;
}
dat <- data.frame(precinct,mat);
names(dat) <- names(dlist[[1]]);
names(dat)[-(1:2)] <-
  c("Greene E Day","Greene Total Votes","Rawl E Day","Rawl Total Votes","Total");
dim(dat);
names(dat);

runtest <- function(wrk, digit=2) {
  getdigits <- function(num,digit=2) {
    s <- as.character(num);
    idx <- sapply(s, function(x){ nchar(x) >= digit; });
    dout <- ifelse(idx, NA, "");
    if (any(idx)) {
      dout[idx] <- sapply(s[idx], function(x){ substr(x,digit,digit) });
    }
    return(dout);
  }
  
  # Benford's Law probablity formulas
  # source
  # http://www.mathpages.com/home/kmath302/kmath302.htm
  #
  # PndB:  d, digit value (vector);  n, digit place;  B, base
  # probability that the n-th digit following the first nonzero digit is d,
  #   for numbers in base B
  # d can be a vector;  n and B must be positive scalars
  #
  PndB <- function(d,n,B=10) {
    if (n==0) {
      p <- log(1+1/d)/log(B);
    }
    else {
      seq <- B^(n-1):(B^n-1);
      p <- d*0;
      for (i in 1:length(d)) {
        p[i] <- sum(log(1+1/(d[i]+seq*B)))/log(B);
      }
    }
    return(p);
  }

  least <- ifelse(digit==1, 1, 0);  # initial digit for Benford test
  vec <- least:9;
  set <- rep(0,length(vec));
  names(set) <- as.character(vec);
  ndtable <- names(dtable <- table(d <- getdigits(wrk, digit=digit)));
  m <- mean(d <- as.numeric(d), na.rm=TRUE);
  sd <- sqrt(var(d, na.rm=TRUE))
  set[ndtable[ndtable %in% vec]] <- dtable[ndtable[ndtable %in% vec]];
  sumset <- sum(set);
  dmean <- sum(set * 0:9)/sum(set);
  # Benford's law logarithmic distribution test
  bpred <- sumset*PndB(vec, digit-1);
  chi <- sum(dev <- (set - bpred)^2 / bpred);
  if(is.na(chi)) print( table(getdigits(wrk, digit=digit)) );
  return(list(N=sumset, chi=chi, mean=m, se=sd/sqrt(sumset)));
}

idxT <- asc(dat$Precinct) %in% c("Totals:");
sum(idxT);
idxTA <- asc(dat$Precinct) %in% c("Totals:","Absentee");
sum(idxTA);
idxA <- asc(dat$Precinct) %in% c("Absentee");
sum(idxA);

# all ballots
lapply(dat[!idxT,c(2,4,6)], runtest);
# 95% and 90% confidence intervals
lapply(lapply(dat[!idxT,c(2,4,6)], runtest),
  function(x){ c(x$mean+x$se*qnorm(1-.025)*c(-1,1),
    x$mean+x$se*qnorm(1-.05)*c(-1,1)) });

# all non-absentee ballots
lapply(dat[!idxTA,c(2,4,6)], runtest);
# 95% and 90% confidence intervals
lapply(lapply(dat[!idxTA,c(2,4,6)], runtest),
  function(x){ c(x$mean+x$se*qnorm(1-.025)*c(-1,1),
    x$mean+x$se*qnorm(1-.05)*c(-1,1)) });

# absentee ballots only
lapply(dat[idxA,c(2,4,6)], runtest);
# 95% and 90% confidence intervals
lapply(lapply(dat[idxA,c(2,4,6)], runtest),
  function(x){ c(x$mean+x$se*qnorm(1-.025)*c(-1,1),
    x$mean+x$se*qnorm(1-.05)*c(-1,1)) });
